@@ -2,7 +2,7 @@
 
 ;; Author: Takafumi Arakaki <aka.tkf at gmail.com>
 ;; Package-Requires: ((emacs "24") (epc "0.1.0") (python-environment "0.0.2") (cl-lib "0.5"))
-;; Version: 0.2.7
+;; Version: 0.2.8
 
 ;; This file is NOT part of GNU Emacs.
 
@@ -41,7 +41,7 @@
   :group 'completion
   :prefix "jedi:")
 
-(defconst jedi:version "0.2.7")
+(defconst jedi:version "0.2.8")
 
 (defvar jedi:source-dir (if load-file-name
                             (file-name-directory load-file-name)
@@ -68,7 +68,10 @@ If it is nil, `python-environment-default-root-name' is used.
 You can specify a full path instead of a name (relative path).
 In that case, `python-environment-directory' is ignored and
 Python virtual environment is created at the specified path."
-  :group 'jedi)
+  :group 'jedi
+  :type '(choice (directory
+                  (const nil))))
+
 
 (defcustom jedi:environment-virtualenv nil
   "``virtualenv`` command to use.  A list of string.
@@ -76,7 +79,9 @@ If it is nil, `python-environment-virtualenv' is used instead.
 
 You must set non-`nil' value to `jedi:environment-root' in order
 to make this setting work."
-  :group 'jedi)
+  :group 'jedi
+  :type '(repeat string))
+
 
 (defun jedi:-env-server-command ()
   (let* ((getbin (lambda (x) (python-environment-bin x jedi:environment-root)))
@@ -127,7 +132,9 @@ reinstall jediepcserver.py.
 If you want to pass some arguments to the Jedi server command,
 use `jedi:server-args' instead of appending them
 `jedi:server-command'."
-  :group 'jedi)
+  :group 'jedi
+  :type '(repeat string))
+
 
 (defcustom jedi:server-args nil
   "Command line arguments to be appended to `jedi:server-command'.
@@ -178,30 +185,40 @@ for Jedi.el and other EPC applications.
 
 If you want to start a new ad-hoc server for the current buffer,
 use the command `jedi:start-dedicated-server'."
-  :group 'jedi)
+  :group 'jedi
+  :type '(repeat string))
+
 
 (defcustom jedi:complete-on-dot nil
   "Non-`nil' means automatically start completion after inserting a dot.
 To make this option work, you need to use `jedi:setup' instead of
 `jedi:ac-setup' to start Jedi."
-  :group 'jedi)
+  :group 'jedi
+  :type 'boolean)
+
 
 (defcustom jedi:tooltip-method '(pos-tip popup)
   "Configuration for `jedi:tooltip-show'.
 This is a list which may contain symbol(s) `pos-tip' and/or
 `popup'.  It determines tooltip method to use.  Setting this
 value to nil means to use minibuffer instead of tooltip."
-  :group 'jedi)
+  :group 'jedi
+  :type '(repeat (choice (const pos-tip)
+                         (const popup))))
 
 (defcustom jedi:get-in-function-call-timeout 3000
   "Cancel request to server for call signature after this period
 specified in in millisecond."
-  :group 'jedi)
+  :group 'jedi
+  :type 'integer)
+
 
 (defcustom  jedi:get-in-function-call-delay 1000
   "How long Jedi should wait before showing call signature
 tooltip in millisecond."
-  :group 'jedi)
+  :group 'jedi
+  :type 'integer)
+
 
 (defcustom jedi:goto-definition-config
   '((nil nil        nil)
@@ -244,26 +261,37 @@ use the setting like this::
 
 You can rearrange the order to have most useful sets of arguments
 at the top."
-  :group 'jedi)
+  :group 'jedi
+  :type '(repeat (choice (const nil)
+                         (const t)
+                         (const definition))))
+
 
 (defcustom jedi:doc-mode 'rst-mode
   "Major mode to use when showing document."
-  :group 'jedi)
+  :group 'jedi
+  :type 'function)
+
 
 (defcustom jedi:doc-hook '(view-mode)
   "The hook that's run after showing a document."
   :type 'hook
   :group 'jedi)
 
-(defcustom jedi:doc-display-buffer 'display-buffer
-  "A function to be called with a buffer to show document."
+(defcustom jedi:goto-definition-hook nil
+  "The hook that's run after jumping to a definition location."
+  :type 'hook
   :group 'jedi)
 
+(defcustom jedi:doc-display-buffer 'display-buffer
+  "A function to be called with a buffer to show document."
+  :group 'jedi
+  :type 'function)
+
 (defcustom jedi:install-imenu nil
-  "[EXPERIMENTAL] If `t', use Jedi to create `imenu' index.
-To use this feature, you need to install the developmental
-version (\"dev\" branch) of Jedi."
-  :group 'jedi)
+  "If `t', use Jedi to create `imenu' index."
+  :group 'jedi
+  :type 'boolean)
 
 (defcustom jedi:imenu-create-index-function 'jedi:create-nested-imenu-index
   "`imenu-create-index-function' for Jedi.el.
@@ -271,7 +299,8 @@ It must be a function that takes no argument and return an object
 described in `imenu--index-alist'.
 This can be set to `jedi:create-flat-imenu-index'.
 Default is `jedi:create-nested-imenu-index'."
-  :group 'jedi)
+  :group 'jedi
+  :type 'function)
 
 (make-obsolete-variable 'jedi:setup-keys nil "0.1.3")
 (defcustom jedi:setup-keys nil
@@ -326,28 +355,53 @@ ropemacs's `rope-show-doc' is same as `jedi:show-doc'.  You can
 avoid collision by something like this::
 
     (setq jedi:key-show-doc (kbd \"C-c D\"))"
-  :group 'jedi)
+  :group 'jedi
+  :type 'boolean)
 
-(defcustom jedi:key-complete (kbd "<C-tab>")
+(defun jedi:-custom-set-key-string (sym defs)
+  "Convert key sequence DEFS from string and store into in SYM."
+  (custom-set-default sym (kbd defs)))
+
+(defun jedi:-custom-get-key-string (sym)
+  "Convert key sequence stored in SYM to string representation."
+  (key-description (default-value sym)))
+
+(defcustom jedi:key-complete "<C-tab>"
   "Keybind for command `jedi:complete'."
-  :group 'jedi)
+  :group 'jedi
+  :type 'string
+  :set #'jedi:-custom-set-key-string
+  :get #'jedi:-custom-get-key-string)
 
-(defcustom jedi:key-goto-definition (kbd "C-.")
+(defcustom jedi:key-goto-definition "C-."
   "Keybind for command `jedi:goto-definition'."
-  :group 'jedi)
+  :group 'jedi
+  :type 'string
+  :set #'jedi:-custom-set-key-string
+  :get #'jedi:-custom-get-key-string)
 
-(defcustom jedi:key-show-doc (kbd "C-c d")
+(defcustom jedi:key-show-doc "C-c d"
   "Keybind for command `jedi:show-doc'."
-  :group 'jedi)
+  :group 'jedi
+  :type 'string
+  :set #'jedi:-custom-set-key-string
+  :get #'jedi:-custom-get-key-string)
 
-(defcustom jedi:key-related-names (kbd "C-c r")
+(defcustom jedi:key-related-names "C-c r"
   "Keybind for command `helm-jedi-related-names' or
 `anything-jedi-related-names'."
-  :group 'jedi)
+  :group 'jedi
+  :type 'string
+  :set #'jedi:-custom-set-key-string
+  :get #'jedi:-custom-get-key-string)
 
-(defcustom jedi:key-goto-definition-pop-marker (kbd "C-,")
+(defcustom jedi:key-goto-definition-pop-marker "C-,"
   "Keybind for command `jedi:goto-definition-pop-marker'."
-  :group 'jedi)
+  :group 'jedi
+  :type 'string
+  :set #'jedi:-custom-set-key-string
+  :get #'jedi:-custom-get-key-string)
+
 
 (defcustom jedi:use-shortcuts nil
   "If non-`nil', enable the following shortcuts:
@@ -355,24 +409,27 @@ avoid collision by something like this::
 | ``M-.``  `jedi:goto-definition'
 | ``M-,``  `jedi:goto-definition-pop-marker'
 "
-  :group 'jedi)
+  :group 'jedi
+  :type 'boolean)
 
 (defcustom jedi:import-python-el-settings t
-  "Automatically import setting from python.el variables."
-  :group 'jedi)
+  "Automatically import settings from python.el variables."
+  :group 'jedi
+  :type 'boolean)
 
 (defcustom jedi:goto-definition-marker-ring-length 16
   "Length of marker ring to store `jedi:goto-definition' call positions"
-  :group 'jedi)
+  :group 'jedi
+  :type 'integer)
 
 
 ;;; Internal variables
 
 (defvar jedi:get-in-function-call--d nil
-  "Bounded to deferred object while requesting get-in-function-call.")
+  "Bound to deferred object while requesting get-in-function-call")
 
 (defvar jedi:defined-names--singleton-d nil
-  "Bounded to deferred object while requesting defined_names.")
+  "Bound to deferred object while requesting defined_names")
 
 
 ;;; Jedi mode
@@ -457,17 +514,64 @@ connection."
        (when something-happened
          ,@run-on-error))))
 
+(defun jedi:epc--make-server-output-msg (line-count)
+  "Extract up to LINE-COUNT last lines from last failed EPC buffer."
+  (let ((epc-buffer (get-buffer (epc:server-buffer-name epc:uid)))
+        is-part lines)
+    (if (not (buffer-live-p epc-buffer))
+        ""
+      (with-current-buffer epc-buffer
+        (save-excursion
+          (goto-char (point-max))
+          (forward-line -10)
+          (setq is-part (not (eq (point) (point-min))))
+          (setq lines (buffer-substring-no-properties (point) (point-max)))))
+      (format "*** EPC Server Output (last %s lines) ***\n%s%s\n"
+              line-count
+              (if is-part "<...snipped, see all with `epc:pop-to-last-server-process-buffer'...>\n" "")
+              lines))))
+
+
 (defun jedi:epc--start-epc (server-prog server-args)
   "Same as `epc:start-epc', but set query-on-exit flag for
 associated processes to nil."
-  (let ((mngr (jedi:-with-run-on-error
-                  (epc:start-epc server-prog server-args)
-                (display-warning 'jedi "\
+  (let ((mngr
+         (condition-case epc-error
+             (epc:start-epc server-prog server-args)
+           (error
+            (let* ((cmdline-args (append (list server-prog) server-args))
+                   (cmd (executable-find server-prog))
+                   (cmd-str (or cmd
+                                (format "nil (%S not found in exec-path)"
+                                        server-prog)))
+                   (virtual-env (getenv "VIRTUAL_ENV"))
+                   (warning-msg (format "
+================================
 Failed to start Jedi EPC server.
+================================
+
+*** EPC Error ***
+%s
+
+%s
+*** EPC Server Config ***
+Server arguments: %S
+Actual command: %s
+VIRTUAL_ENV envvar: %S
+
+*** jedi-mode is disabled in %S ***
+Fix the problem and re-enable it.
+
 *** You may need to run \"M-x jedi:install-server\". ***
 This could solve the problem especially if you haven't run the command yet
 since Jedi.el installation or update and if the server complains about
-Python module imports." :error))))
+Python module imports."
+                                        (cadr epc-error)
+                                        (jedi:epc--make-server-output-msg 10)
+                                        cmdline-args cmd-str virtual-env
+                                        (current-buffer))))
+              (display-warning 'jedi warning-msg :error)
+              (jedi-mode 0))))))
     (set-process-query-on-exit-flag (epc:connection-process
                                      (epc:manager-connection mngr))
                                     nil)
@@ -480,15 +584,23 @@ Python module imports." :error))))
 (defvar jedi:server-pool--table (make-hash-table :test 'equal)
   "A hash table that holds a pool of EPC server instances.")
 
+(defun jedi:server-pool--resolve-command (command)
+  "Resolve COMMAND using current environment.
+Tries to find (car command) in \"exec-path\"."
+  (let ((command-path (executable-find (car command))))
+    (if command-path
+        (cons command-path (cdr command))
+      command)))
+
 (defun jedi:server-pool--start (command)
-  "Get an EPC server instance from server pool by COMMAND as a
-key, or start new one if there is none."
-  (let ((cached (gethash command jedi:server-pool--table)))
+  "Get an EPC server for COMMAND from server pool or start a new one."
+  (let* ((resolved-command (jedi:server-pool--resolve-command command))
+         (cached (gethash resolved-command jedi:server-pool--table)))
     (if (and cached (jedi:epc--live-p cached))
         cached
       (let* ((default-directory "/")
-             (mngr (jedi:epc--start-epc (car command) (cdr command))))
-        (puthash command mngr jedi:server-pool--table)
+             (mngr (jedi:epc--start-epc (car resolved-command) (cdr command))))
+        (puthash resolved-command mngr jedi:server-pool--table)
         (jedi:server-pool--gc-when-idle)
         mngr))))
 
@@ -546,7 +658,22 @@ later when it is needed."
   (setq jedi:get-in-function-call--d nil)
   (setq jedi:defined-names--singleton-d nil))
 
+(defun jedi:stop-all-servers ()
+  "Stop all live Jedi servers.
+This is useful to apply new settings or VIRTUAL_ENV variable
+value to all buffers."
+  (interactive)
+  ;; Kill all servers attached to buffers
+  (cl-dolist (buf (buffer-list))
+    (when (buffer-live-p buf)
+      (with-current-buffer buf
+        (when (jedi:epc--live-p jedi:epc)
+          (jedi:stop-server)))))
+  ;; Kill all unused servers too.
+  (jedi:server-pool--gc))
+
 (defun jedi:get-epc ()
+  "Get an EPC instance of a running server or start a new one."
   (if (jedi:epc--live-p jedi:epc)
       jedi:epc
     (jedi:start-server)))
@@ -589,7 +716,8 @@ See: https://github.com/tkf/emacs-jedi/issues/54"
 (defun jedi:call-deferred (method-name)
   "Call ``Script(...).METHOD-NAME`` and return a deferred object."
   (let ((source      (buffer-substring-no-properties (point-min) (point-max)))
-        (line        (count-lines (point-min) (min (1+ (point)) (point-max))))
+        ;; line=0 is an error for jedi, but is possible for empty buffers.
+        (line        (max 1 (count-lines (point-min) (min (1+ (point)) (point-max)))))
         (column      (- (point) (line-beginning-position)))
         (source-path (jedi:-buffer-file-name)))
     (epc:call-deferred (jedi:get-epc)
@@ -770,6 +898,16 @@ INDEX-th result."
   (forward-line (1- line))
   (forward-char column))
 
+(defvar jedi:find-file-function #'jedi:find-file
+  "Function that reads a source file for jedi navigation.
+It must take these arguments: (file-to-read other-window-flag line_number column_number).")
+
+(defun jedi:find-file (file line column other-window)
+  "Default function used by jedi to find a source FILE, LINE= and COLUMN, possibly in OTHER-WINDOW."
+  (funcall (if other-window #'find-file-other-window #'find-file)
+           file)
+  (jedi:goto--line-column line column))
+
 (defun jedi:goto-definition--nth (other-window &optional try-next)
   (let* ((len (length jedi:goto-definition--cache))
          (n jedi:goto-definition--index)
@@ -790,9 +928,8 @@ INDEX-th result."
           (message "File '%s' does not exist." module_path)))
        (t
         (jedi:goto-definition-push-marker)
-        (funcall (if other-window #'find-file-other-window #'find-file)
-                 module_path)
-        (jedi:goto--line-column line_nr column)
+        (jedi:find-file module_path line_nr column other-window)
+        (run-hooks 'jedi:goto-definition-hook)
         (jedi:goto-definition--notify-alternatives len n))))))
 
 (defun jedi:goto-definition--notify-alternatives (len n)
@@ -1172,6 +1309,16 @@ See also:
       (lambda (_)
         (setq-default jedi:server-command (jedi:-env-server-command))))))
 
+;;;###autoload
+(defun jedi:reinstall-server ()
+  "Reinstall Jedi server script jediepcserver.py."
+  (interactive)
+  (let ((root (python-environment-root-path jedi:environment-root)))
+    (when (yes-or-no-p
+           (format "Reinstall jedi server environment(%s) ?" (abbreviate-file-name root)))
+      (delete-directory root t)
+      (jedi:install-server))))
+
 ;;;###autoload
 (defun jedi:install-server-block ()
   "Blocking version `jedi:install-server'."
@@ -1183,9 +1330,11 @@ See also:
 
 (defcustom jedi:install-python-jedi-dev-command
   '("pip" "install" "--upgrade"
-    "git+https://github.com/davidhalter/jedi.git@dev#egg=jedi")
+    "git+https://github.com/davidhalter/jedi.git@master#egg=jedi")
   "Pip command to be used for `jedi:install-python-jedi-dev'."
-  :group 'jedi)
+  :group 'jedi
+  :type '(repeat string)
+  )
 
 (defun jedi:install-python-jedi-dev ()
   "Install developmental version of Python-Jedi from GitHub."
